from os import scandir
from collections import namedtuple

PortDef = namedtuple("PortDef", "udp tcp")
ServiceDef = namedtuple("ServiceDef", "title description ports categories reference")

def split_port_def(port_def):
    ports_dict = {"udp": [], "tcp": []}

    # In the worst case, port_def is "13:15,124/udp|120:122"
    # Which should mean 13 to 15 and 124, all udp, and 120 to 122 both udp and tcp
    try:
        proto_subdefs = port_def.split("|")
    except ValueError:
        proto_subdefs = (port_def, )

    # first port_subdef is "13:15,124,128/udp"
    for proto_subdef in proto_subdefs:
        try:
            port_subdef, proto_def = proto_subdef.split("/")
            if proto_def == "tcp6":
                protos = ("tcp", )
                print("  WARNING : using tcp for tcp6")
            else:
                protos = (proto_def, )
        except ValueError:
            port_subdef = proto_subdef
            protos = ("udp", "tcp")

        # first port_subdef is now "13:15,124,128" : protos is ("udp", )
        try:
            port_list = port_subdef.split(",")
        except ValueError:
            port_list = [ port_subdef ]

        # port_list is ("13:15", "124", "128"). Port ranges can be very wide,
        # yielding 10000 entries. We'll keep them as ranges.
        for proto in protos:
            for port in port_list:
                ports_dict[proto].append(port)

    return PortDef(ports_dict["udp"], ports_dict["tcp"])


def scan_service_definitions(definitions_dir):
    '''scan service definitions in the list given by jhansonxi'''
    # Files in the definitions dir define one or more protocols and
    # their ports. Each file comes with the following format :
    #
    # [NAME_ONE]
    # title=EXPANDED NAME
    # description=BLA BLA BLA
    # categories=ONE;TWO;
    # reference=[HTTP://HTML_LINK LINK_NAME DESCRIPTION]
    # ports=220|3724,6112:6114,4000/tcp
    #
    # [NAME_TWO]
    # etc...

    filelist = scandir(definitions_dir)
    services = {}
    for filename in iter(filelist):
        print("processing %s" % filename.path)
        with open(filename.path, "r") as fd:
            # NAME_ONE and NAME_TWO are separated by a double newline
            service_list = fd.read().split("\n\n")
            for service in service_list:
                # _Some_ service descriptions finish in "\n\n"
                if service == "" or service == "\n":
                    continue
                service_desc = service.split("\n")
                service_name = service_desc[0].strip("[]")
                print("  -> %s" % service_name)

                # get rid of empty strings :
                for i in range(service_desc.count("")):
                    service_desc.remove("")

                # services['NAME_ONE'] will return a dict with its parameters :
                service_def = {
                    key: value for (key, value) in
                        ( item.split("=", 1) for item in iter(service_desc[1:]) )
                }
                if not "reference" in service_def:
                    service_def["reference"] = ""
                service_def["ports"] = split_port_def(service_def["ports"])
                s = ServiceDef(
                        service_def["title"],
                        service_def["description"],
                        service_def["ports"],
                        service_def["categories"],
                        service_def["reference"],
                )
                services[service_name] = s

    return services




# Making this directory a python package.
name = "servicewall"
__all__ = [
    "service_helpers",
    "network_helpers",
    "main",
    "statefulfirewall",
    "firewall",
]
#from pkgutil import extend_path
#__path__ = extend_path(__path__, __name__)

# Load the main package
from servicewall.main import ServiceWall

# This function needs to be here and have service_helpers imported
# this way to have working pickle in the ServiceWall class.
from servicewall import service_helpers
def update_service_defs():
    import pickle
    service_pickle = "lib/services.p"
    service_defs_dir = "/etc/gufw/app_profiles"
    s = service_helpers.scan_service_definitions(service_defs_dir)
    print("writing defs from %s to %s" % (service_defs_dir, service_pickle))
    if os.path.isfile(service_pickle):
        raise SystemError("There's already a file named %s" % service_pickle)
    with open(service_pickle, "wb") as fd:
        pickle.dump(s, fd)




"""Firewall class

implements a firewall using python-iptables
"""

from os import environ
from iptc import Rule, Match, Chain, Table
from iptc.ip4tc import IPTCError



class FireWall():
    """A simple firewall class

    In this firewall, there is (only) one _table called FILTER. It contains
    2 chains called input and output. When anything comes from outside, it
    goes through input. When anything goes outside, it goes through output.

    You can add (and delete) rules based on any valid iptables match.

    TODO :
    - ruleset Object.
    """

    identifier = "ServiceWall"

    def __init__(self):
        try:
            self._table = Table(Table.FILTER)
        except IPTCError:
            # user is not root - return silently ; all writing ops will fail
            return

        # _table.autocommit is True by default.
        #self._table.autocommit = False
        self.input_chain = Chain(self._table, "INPUT")
        self.forward_chain = Chain(self._table, "FORWARD")
        self.output_chain = Chain(self._table, "OUTPUT")
        if self.status():
            self.up = True
        else:
            self.up = False

    def start(self):
        """Start a basic FireWall.

        Drops all incoming, accepts only localhost.
        """
        # Drop all incoming. Basic but functional.
        print("setting input policy to DROP")
        self.input_chain.set_policy("DROP")
        print("setting forward policy to DROP")
        self.forward_chain.set_policy("DROP")
        print("setting output policy to ACCEPT")
        self.output_chain.set_policy("ACCEPT")
        # Accept all on localhost.
        self.add_rule(
                "localhost",
                self.input_chain,
                "ACCEPT",
                iface="lo"
                )
        if not self._table.autocommit:
            self._table.commit()
        self.up = True

    def stop(self):
        """Remove rules.

        This function removes all rules created by this firewall's instances ;
        it recognizes them by the comment that's inside of them.
        """
        print("flushing rules")
        found_rules = 0
        for rule in self.input_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                self.del_rule(self._get_rule_name(rule), self.input_chain)
                found_rules += 1
        for rule in self.forward_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                self.del_rule(self._get_rule_name(rule), self.forward_chain)
                found_rules += 1
        if not found_rules:
            print("no rule found for id %s." % self.identifier)
        print("setting input policy to ACCEPT")
        self.input_chain.set_policy("ACCEPT")
        print("setting forward policy to ACCEPT")
        self.forward_chain.set_policy("ACCEPT")
        if not self._table.autocommit:
            self._table.commit()
        self.up = False

    def status(self):
        """print the status of the FireWall. Returns either True or False."""
        # Returns True if a single rule has our identifier tag
        for rule in self.input_chain.rules:
            if self._get_rule_id(rule) == self.identifier:
                return True
        else:
            return False

    def add_rule(self, name, chain, target, dst="", dport="", src="", sport="", proto="", iface="", position="top"):
        """Add a rule to the input chain self.input_chain.

        chain should be either self.input_chain or self.output_chain.
        name can be any string
        target is a string saying what to do, like
                "ACCEPT", "DENY", "DROP", "LOG"
        proto (if not defined, is assumed both udp and tcp)
        src is a realm in the form XXX.XXX.XXX.XXX/YY
        sport is source port. Currently ignored
        dst is a realm in the same form as src
        dport is destination port
        iface is the interface, will be in_interface if chain is the input chain.
                                       out_interface if chain is the output chain.
        """
        #print("adding rule %s" % name)
        if (dport or sport) and not proto:
            for proto in "tcp", "udp":
                self.add_rule(name, chain, target, dst, dport, src, sport, proto, iface)
        else:
            rule = Rule()
            rule.create_target(target)
            if target == "LOG":
                limit_match = rule.create_match("limit")
                limit_match.limit = "1/s"
                limit_match.limit_burst = "1"
                rule.target.set_parameter("log-prefix", self.identifier + ":")
            # First we need to know if we go in or out
            # Then some rules need the creation of a match
            if dst:
                rule.dst = dst
            if dport:
                proto_match = rule.create_match(proto)
                proto_match.dport = str(dport)
                rule.protocol = proto
            if src:
                rule.src = src
            if sport:
                proto_match = rule.create_match(proto)
                proto_match.sport = str(sport)
                rule.protocol = proto
            if iface:
                if chain.name == self.input_chain.name:
                    rule.in_interface = iface
                elif chain.name == self.output_chain.name:
                    rule.out_interface = iface
                if src and src == dst:
                    print("warning, both source and dest are set to %s" % src)
           # Add a signature as a comment.
            comment_match = rule.create_match("comment")
            comment_match.comment = self.identifier + ":" + name
            # Try to set it into chain.
            rule.final_check()
            if position == "bottom":
                chain.append_rule(rule)
            elif position == "top":
                chain.insert_rule(rule)
            else:
                raise SystemError("position should be top or bottom")
            if not rule in chain.rules:
                print("Need to check %s rule." % name)

    def del_rule(self, name, chain):
        """Remove rule.

        This function removes one rule named "name" in chain "chain". It
        expects the rule to contain a comment in the form
            "identifier:name"
        where identifier is the name of the program that has set the rule, and
        title is the name of the rule. Note that there may be several rules
        under the same name. This should delete the first it finds.
        """
        for rule in chain.rules:
            if self._get_rule_name(rule) == name:
                chain.delete_rule(rule)
                #print("deleted rule %s" % name)
                break

    def list_rules(self, chain):
        for rule in chain.rules:
            # First find out dport - it gets set by self.add_service_in as a match.
            dport = "any port"
            proto = "tcp/udp"
            for match in rule.matches:
                try:
                    dport = "port " + match.parameters["dport"]
                    proto = match.name
                except (KeyError, IndexError):
                    pass
            # Then find out source.
            if rule.src == "0.0.0.0/0.0.0.0":
                src = "any source"
            else:
                src = rule.src
            yield "%6s %-20s : %7s %-8s from %-23s" % (
                rule.target.name, self._get_rule_name(rule), proto, dport, src
            )

    def _get_rule_name(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    rule_identifier, rule_name = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_name

    def _get_rule_id(self, rule):
        for match in rule.matches:
            if match.comment:
                try:
                    rule_identifier, rule_name = match.comment.split(":")
                except IndexError:
                    # No semicolon in comment
                    continue
                return rule_identifier




#!/usr/bin/env python
# -*- coding: utf-8 -*-


import statefulfirewall


fw = statefulfirewall.StateFulFireWall()
l = []

print(fw.protobynumber)
for i in fw.yield_logs(limit=100):
        l.append(i)
for log in l:
    if 'conntrack_type' in log:
        #print('%-8s %s' % (log['conntrack_type'], log['id']))
        continue
    else:
        if 'DPT' not in log:
            log['DPT'] = ''
        if log['PROTO'].isnumeric():
            log['PROTO'] = fw.protobynumber[int(log['PROTO'])]
        print('packet   %-16s %-16s %4s %5s %-15s' % (
                    log['SRC'],
                    log['DST'],
                    log['PROTO'],
                    log['DPT'],
                    str(log['LOG_DATE']),
        ))




#!/usr/bin/env python


import os
import wx.adv
import wx
import json
import servicewall


lib_path = "/usr/lib/servicewall/"
#lib_path = os.path.dirname(os.path.abspath(__file__)) + "/"
LOG_LIMIT = 10
TRAY_TOOLTIP = 'ServiceWall' 
TRAY_ICON = lib_path + "icon.png"
TRAY_ICON2 = lib_path + "icon2.png"


class TaskBarIcon(wx.adv.TaskBarIcon):
    """A systray icon with a panel showing latest hits
    """
    def __init__(self, frame):
        self.frame = frame
        super(TaskBarIcon, self).__init__()
        self.SetIcon(wx.Icon(wx.IconLocation(TRAY_ICON2)), "ServiceWall")
        self.Bind(wx.adv.EVT_TASKBAR_LEFT_DOWN, self.on_left_down)

    def CreatePopupMenu(self):
        """gets automatically called on systray's right-click"""
        menu = wx.Menu()
        exit_item = menu.Append(-1, "Exit", "exit")
        self.Bind(wx.EVT_MENU, self.on_exit, exit_item)
        #menu.AppendSeparator()
        return menu

    def CreateApplet(self):
        r = wx.GetClientDisplayRect()
        position = (r.top, r.right)
        window = wx.PopupTransientWindow(self.frame, flags=wx.BORDER_NONE)
        fw = servicewall.ServiceWall()
        yielder = fw.yield_logs(limit=LOG_LIMIT)
        panel = wx.Panel(window)
        i = 0
        panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        icons = []
        labels = []
        logs_by_port = {}

        # First sort logs by port :
        for log in yielder:
            if log["DPT"] not in logs_by_port:
                logs_by_port[log["DPT"]] = [log,]
            else:
                logs_by_port[log["DPT"]].append(log)

        # Then display them :
        status_sizer = wx.BoxSizer(wx.VERTICAL)
        logs_sizer = wx.GridSizer(len(logs_by_port) + 1, 4, 0, 0)
        status_label = wx.StaticText(panel, label="", style=wx.ALIGN_CENTER)
        status_details = wx.StaticText(panel, label="\nrealm : %s" % fw.realm_id, style=wx.ALIGN_CENTER)
        if fw.config["enabled"]:
            status_str = "enabled"
        else:
            status_str = "disabled"
        status_label.SetLabelMarkup("<big>ServiceWall\n%s</big>" %
                (status_str))
        status_sizer.Add(status_label, 1, wx.ALIGN_CENTER, 25)
        status_sizer.Add(status_details, 1, wx.ALIGN_CENTER, 25)
        panel_sizer.Add(status_sizer, 1, wx.ALIGN_CENTER, 5)
        panel_sizer.Add(logs_sizer, 1, wx.ALL, 20)
        image = wx.Image(TRAY_ICON2, type=wx.BITMAP_TYPE_ANY).Scale(32, 32)
        bitmap = wx.Bitmap(image)
        logs_sizer.Add(wx.StaticText(panel, label=""))
        logs_sizer.Add(wx.StaticText(panel, label='port'), 1, wx.ALIGN_RIGHT, 0)
        logs_sizer.Add(wx.StaticText(panel, label="hits"), 1, wx.ALIGN_RIGHT, 0)
        logs_sizer.Add(wx.StaticText(panel, label="age"), 1, wx.ALIGN_RIGHT, 0)
        for name, logpile in logs_by_port.items():
            date = logpile[0]["LOG_DATE"]
            delta = date.now() - date
            if delta.seconds <= 60:
                age = str(delta.seconds) + "'"
            else:
                age = ":".join(str(delta).split(".")[0].split(":")[0:2])
            #rawbitmap = wx.Bitmap(TRAY_ICON2, type=wx.BITMAP_TYPE_ANY)
            #bitmap = wx.Bitmap(rawbitmap.ConvertToImage().Rescale(40, 40))
            icons.append(wx.StaticBitmap(panel, -1, bitmap))
            labels.append([
                    wx.StaticText(panel, label="%s" % logpile[0]["DPT"], style=wx.ALIGN_RIGHT),
                    wx.StaticText(panel, label="%i" % len(logpile)),
                    wx.StaticText(panel, label="%s" % age)
            ])
            logs_sizer.Add(icons[i], 1, wx.ALL, 0)
            logs_sizer.Add(labels[i][0], 1, wx.ALIGN_RIGHT, 25)
            logs_sizer.Add(labels[i][1], 1, wx.ALIGN_RIGHT, 0)
            logs_sizer.Add(labels[i][2], 1, wx.ALIGN_RIGHT, 0)
            #for j in range(3):
            #    logs_sizer.Add(labels[i][j], 6)
            i += 1
        panel.SetSizer(panel_sizer)
        panel.SetAutoLayout(True)
        panel_sizer.Fit(panel)
        window.SetPosition((r.right - panel_sizer.Size[0] + 1, r.top))
        window.SetSize(panel_sizer.Size)
        #window.Show(True)
        window.Popup()
        return window

    def on_left_down(self, event):      
        #self.set_icon(TRAY_ICON2)
        #self.PopupMenu(self.CreatePopupMenu())
        tray = event.GetEventObject()
        self.CreateApplet()

    def on_hello(self, event):
        print('Hello, world!')

    def log_callback(self, event):
        menu = event.GetEventObject()
        m = menu.GetMenuItems()[event.Id]

    def on_exit(self, event):
        wx.CallAfter(self.Destroy)
        self.frame.Close()


class App(wx.App):
    def OnInit(self):
        frame = wx.Frame(None)
        self.SetTopWindow(frame)
        TaskBarIcon(frame)
        return True


def main():
    app = App(False)
    app.MainLoop()

if __name__ == '__main__':
    main()




"""  StateFulFireWall

Uses simple guidelines given in

and implements them in a FireWall class, using reasonable defaults.
"""

from systemd import journal
from iptc import Rule
from servicewall import firewall
from datetime import datetime
import socket


class StateFulFireWall(firewall.FireWall):
    """Implement some useful stateful rules :
    
    - accept related/established packets
    - log anything that is dropped
    - drop invalid packets
    """
    protobynumber = { num: name[8:].lower()
                 for name, num in vars(socket).items()
                 if name.startswith("IPPROTO")
    }
    def __init__(self):
        super().__init__()

    def start(self, **args):
        print("adding rule icmp")
        rule = Rule()
        rule.create_target("ACCEPT")
        rule.protocol = "icmp"
        m = rule.create_match("icmp")
        m.set_parameter("icmp-type", "8")
        comment_match = rule.create_match("comment")
        comment_match.comment = self.identifier + ":icmp"
        self.input_chain.append_rule(rule)

        # Top rule should be to allow related, established connections.
        self.add_conntrack_rule_in("ACCEPT", "ctstate", "RELATED,ESTABLISHED")

        # Log all that is refused.
        self.add_rule(
                "nflog",
                self.input_chain,
                "NFLOG",
                position="bottom",
        )
        self.add_rule(
                "nflog",
                self.forward_chain,
                "NFLOG",
                position="bottom",
        )
 
        # Drop invalid packets - as diagnosed by the conntrack processor.
        self.add_conntrack_rule_in("DROP", "ctstate", "INVALID", position="bottom")

        super().start(**args)   # commits the table if relevant

    def add_conntrack_rule_in(self, target, param_key, param_value, position="top"):
        """Adds a rule to input chain, for example :

        StateFulFireWall.add_conntrack_rule_in("ACCEPT", "ctstate", "RELATED")
        """
        print("adding rule %s" % param_value)
        conntrack_rule = Rule()
        conntrack_rule.create_target(target)
        conntrack_match = conntrack_rule.create_match("conntrack")
        conntrack_match.set_parameter(param_key, param_value)
        comment_match = conntrack_rule.create_match("comment")
        comment_match.comment = self.identifier + ":" + param_value
        if position == "top":
            self.input_chain.insert_rule(conntrack_rule)
        elif position == "bottom":
            self.input_chain.append_rule(conntrack_rule)


    def yield_logs(self, limit=None, period=None):
        """get logs we implemented in iptables from journald

        limit [int] is the max number of logs to yield
        period [int] is the age in seconds of the oldest log to yield
        """
        reader = journal.Reader()
        #reader.log_level(journal.LOG_WARNING)
        # That would be for LOG match :
        #reader.add_match(SYSLOG_IDENTIFIER="kernel")
        # That is for ulog systemd service associated to ulog.socket :
        reader.add_match(_SYSTEMD_UNIT="servicewall-logs.service")
        now = datetime.now()
        hostname = socket.gethostname()
        #p = select.poll()
        #p.register(reader, reader.get_events())
        #p.poll()
        reader.seek_tail()
        if limit:
            limit = int(limit)
            i = 1
        if period:
            period = int(period)
        while True:
            log = reader.get_previous()
            if not "MESSAGE" in log:
                continue
            # Quit if log older than period :
            if period:
                age = now - log["__REALTIME_TIMESTAMP"]
                if int(age.total_seconds()) > period:
                    break
            message_dict = {}
            # Get rid of the trailing date and hostname :
            try:
                message = log["MESSAGE"].split(hostname)[-1].strip()
            except IndexError:
                print("!! error :")
                print(log["MESSAGE"])

            # If message begins with "[NEW] "or "[DESTROY]", then it is a
            # conntrack info log - skip.
            if message.startswith("[NEW]") or message.startswith("[DESTROY]"):
                continue

            # Only count the logs that make it here :
            if limit:
                if i > limit:
                    break
                else:
                    i += 1

            message_dict["LOG_DATE"] = log["__REALTIME_TIMESTAMP"]
            for item in message.split():
                if item.count("="):
                    key, value = item.split("=")
                    message_dict[key] = value
                else:
                    message_dict[item] = ""

            if 'DPT' not in message_dict:
                message_dict['DPT'] = ''
            if 'SPT' not in message_dict:
                message_dict['SPT'] = ''
            if message_dict['PROTO'].isnumeric():
                message_dict['PROTO'] = self.protobynumber[int(message_dict['PROTO'])]
            else:
                message_dict['PROTO'] = message_dict['PROTO'].lower()

            yield message_dict

    def filter_logs_by(self, criteria, limit=None, period=None):
        """output logs sorted by a log variable, like "DPT" or "SRC"

        note that limit applies to the criteria (like sort nth first SRC hosts)
        """
        yielder = self.yield_logs(period=period)
        logs = {}
        if limit:
            limit = int(limit)
            i = 0
        if period:
            period = int(period)
        for log in yielder:
            if log[criteria] not in logs:
                logs[log[criteria]] = [log, ]
                if limit:
                    if i >= limit:
                        # Stop at step i+1 and pop last record :
                        logs.popitem()
                        break
                    else:
                        i += 1
            else:
                logs[log[criteria]].append(log)
        return logs

